#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <network.h>
#include <ogcsys.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/iosupport.h>
#include <sdcard/wiisd_io.h>
#include <ogc/usbstorage.h>
#include <smb.h>
#include <errno.h>
#include <ogc/mutex.h>
#include "file.h"
#include "share.h"
#include "common.h"

#define READ_AHEAD_AMOUNT			2048
#define MAX_READ_AHEAD				65536

#define ENUMERATE_DEVICES_STACK     (64 * 1024)
#define ENUMERATE_DEVICES_PRIORITY  63

int File::instance = 0;

// So multiple files can see the same failure
FileSystem *File::filesystems = 0;
uint32_t File::filesystemcount = 0;

// For shares
ShareList *File::shares = 0;
uint32_t File::sharecount = 0;

#ifdef DEBUG_MODE
int File::readAheadAmount = 0;
int File::readAheadMax = 0;
#endif

mutex_t readAheadLock;
mutex_t detectLock;

// Disc interface definitions
DISC_INTERFACE* sd = (DISC_INTERFACE*)&__io_wiisd;
DISC_INTERFACE* usb = (DISC_INTERFACE*)&__io_usbstorage;

void *enumerateDevicesThread( void * args )
{
    File file;
    
    // Enumerate any shares
    file.enumerateShares();
    
    while( 1 )
    {
        // Simply try to enumerate devices
        file.enumerateInterfaces();
        usleep( 500000 );
    }
    
    return NULL;
}

File::File()
{
	// Start out wih nothing
	fType = STREAM_NONE;
	
	// To ensure reading strings does not error
	last_char = 0;
	
	// Just for safety sake
	fp = NULL;
	dir = NULL;
	
	if(instance == 0)
	{
        // Initialize mutex for reading ahead
        LWP_MutexInit(&readAheadLock, false);
        
        // Initialize mutex for detecting FS changes
        LWP_MutexInit(&detectLock, false);
        
        // Set up the network
        for(int i = 0; i < MAX_INIT_RETRIES && net_init() == -EAGAIN; i++) { ; }
        
        // Enumerate interfaces we have (to get sd initialized)
        enumerateInterfaces();
        
        // Create a thread that checks every so often for changes
        static lwp_t id;
        LWP_CreateThread(&id, enumerateDevicesThread, NULL, NULL, ENUMERATE_DEVICES_STACK, ENUMERATE_DEVICES_PRIORITY);
	}
	
	// Remember count (never subtract because we never want this information above to go away, even if there are no more instances at the moment)
	instance++;
}

File::~File()
{
    // Could have a resources leak if we were to exit to something that didn't ignore our mallocs
    // due to filesystems being dynamically added to
    
	// Close any open handles
	close();
}

void File::addInterface( const char * const protocol, DISC_INTERFACE * const interface )
{
    // We have another!
    filesystemcount++;    
    filesystems = (FileSystem *)realloc( filesystems, sizeof( filesystems[0] ) * filesystemcount );
    
    // Link up!
    filesystems[filesystemcount-1].protocol = strdup( protocol );
    filesystems[filesystemcount-1].interface = interface;
    
    // Ensure sanity is kept
    int len = strlen(filesystems[filesystemcount-1].protocol);
    if( filesystems[filesystemcount-1].protocol[len-1] == ':' )
    {
        // Chop off
        filesystems[filesystemcount-1].protocol[len-1] = 0;
    }
    
    // Default for now
    filesystems[filesystemcount-1].writeable = false;
    filesystems[filesystemcount-1].revision = 0;
    
    // Only mount if we give it an interface
    if( interface )
    {
        if( !(interface->startup() && fatMountSimple(protocol, interface)) )
        {
            // Failed to init, don't do anything
            filesystems[filesystemcount-1].failed = true;
        }
        else
        {
            // Worked!
            filesystems[filesystemcount-1].failed = false;
        }
    }
}

FileSystem *File::findFileSystem( const char * const url )
{
    // So we can chop off the part beginning with :
    FileSystem *ret = 0;
    char *protocol = strdup( url );    
    char *loc = strchr( protocol, ':' );
    
    if( loc )
    {
        // Chop off str
        loc[0] = 0;
    }
    
    for( uint32_t i = 0; i < filesystemcount; i++ )
    {
        if( strcmp( protocol, filesystems[i].protocol ) == 0 )
        {
            ret = &filesystems[i];
        }
    }
    
    // No memleaks please
    free( protocol );
    
    // Return if we found it
    return ret;
}

bool File::hasFailed( const char * const protocol )
{
    bool ret = true;
    FileSystem *system = findFileSystem( protocol );
    
    // Is the failed flag explicitly set?
    if( system )
    {
        ret = system->failed;
    }
    
    // If we have a good filesystem, has it changed under our nose?
    if( ret == false )
    {
        if( revision != getRevision( protocol ) )
        {
            // Filesystem changed
            ret = true;
        }
    }
    
    return ret;
}

void File::setFailed( const char * const protocol )
{
    FileSystem *system = findFileSystem( protocol );
    
    if( system )
    {
        system->failed = true;
    }
}

bool File::isWriteable( const char * const protocol )
{
    bool ret = false;
    FileSystem *system = findFileSystem( protocol );
    
    if( system )
    {
        ret = system->writeable;
    }
    
    return ret;
}

uint32_t File::getRevision( const char * const protocol )
{
    uint32_t ret = 0;
    FileSystem *system = findFileSystem( protocol );
    
    if( system )
    {
        ret = system->revision;
    }
    
    return ret;
}

int File::openfat(const char *fName, const char *mode)
{
	// Check for directory mode
	if(mode[0] == 'd' || mode[0] == 'D')
	{
		// Open as directory
		dir = diropen(fName);
		
		listedBack = false;
		
		if(!dir)
		{
			// Error!
			return -1;
		}
	}
	else
	{
		// Open as file
		fp = fopen(fName, mode);
		
		if(!fp)
		{
			// Error!
			return -1;
		}
	}
    
    // Remember filesystem revision
    revision = getRevision( fName );
	
	return 0;
}

bool File::exists(const char *url, const char *mode)
{
    if(strcmp(url, "/") == 0) // root directory list
	{
        // Always does
        return true;
	}
	else if( findFileSystem(url) ) // file url
	{
        // Try to open
		if(!openfat(url, mode))
		{
			// Success!
            close();
            
			return true;
		}
    }
    
    return false;
}

int File::open(const char *url, const char *mode)
{
	// Set FileType temporarily
	fType = STREAM_NONE;
	
	// Remember filename
	strcpy(currentUrl, url);
	
	// Clear buffer
	rBuf.init();
    
	if(strcmp(url, "/") == 0) // root directory list
	{
		if(mode[0] == 'd' || mode[0] == 'D')
		{
			// Set the mode up to give back root entries
			rootPos = 0;
            
            // Must set after enumerate
            fType = STREAM_ROOT;
            
			return 0;
		}
        else
        {
            // Just in case
            fType = STREAM_ROOT;
        }
		
		return -1;
	}
	else if( findFileSystem(url) ) // file url
	{
		fType = STREAM_FILE;
		
        // Try to open
		if(!openfat(url, mode))
		{
			// Success!
			return 0;
		}
		
        // Set FileType properly to failed
        setFailed(url);
		return -1;
	}
    else
    {
        // Unhandled stream type
    }
	
	// Default is error
	return -1;
}

void File::readAhead()
{
	// Don't do anything if this medium is failed
	if(fType == STREAM_NONE) { return; }
	if(hasFailed( currentUrl )) { return; }
	if(rBuf.amount() >= MAX_READ_AHEAD) { return; }
	
	static uint8_t data[READ_AHEAD_AMOUNT];
    
    // Read a spot ahead
    int readAmount = readInternal(data, READ_AHEAD_AMOUNT);
    
    // Add to ring buffer
    if(readAmount > 0)
    {
        rBuf.put(data, readAmount);
    }

#ifdef DEBUG_MODE
    readAheadMax = MAX_READ_AHEAD;
    readAheadAmount = rBuf.amount();
#endif
}

int File::readInternal(const void *buffer, uint32_t amount)
{
	// Read from current stream
	switch(fType)
	{
		case STREAM_FILE:
		{
            // Ensure we don't step on our own toes reading ahead
            LWP_MutexLock(readAheadLock);
            
			int ret = fread((void *)buffer, 1, amount, fp);
            
            // Toes status: [ ] stepped on [x] not stepped on
            LWP_MutexUnlock(readAheadLock);
			
			if(ret >= 0)
			{
				// Ensure no error
				if(ferror(fp))
				{
					// Ensure we close right
                    setFailed( currentUrl );
					
					// Failure
					return -1;
				}
				
				// Fine
				return ret;
			}
			else
			{
				// Set flag
                setFailed( currentUrl );
				
				// Not fine
				return ret;
			}
			break;
		}
		default:
			// Catch all other cases
			return 0;
	}
}

int File::read(const void *buffer, uint32_t size, uint32_t count)
{
	// Don't do anything if this medium is failed
	if( hasFailed( currentUrl ) ) { return -1; }
	
	// Know the size
	uint32_t amount = size * count;
	uint32_t offset = 0;
	
	// Copy out of ringbuffer if it exists
	if(rBuf.amount() >= amount)
	{
		// Grab out of ringbuffer
		rBuf.get((void *)buffer, amount);
		
		// Return full amount
		return amount;
	}
	
	if(rBuf.amount() > 0)
	{
		// Remember how much and where to read to
		amount -= rBuf.amount();
		offset += rBuf.amount();
		
		// Grab from the buffer
		rBuf.get((void *)buffer, rBuf.amount());
	}
	
	// Read the rest out of memory
	int newAmount = readInternal(((uint8_t *)buffer) + offset, amount);
	
	if(newAmount > 0)
	{
		// Return the actual count read
		return (offset + newAmount) / size;
	}
	else
	{
		// Return only the first part
		return offset / size;
	}
}

int File::write(const void *buffer, uint32_t size, uint32_t count)
{
	// Don't do anything if this medium is failed
	if( hasFailed( currentUrl ) ) { return -1; }
	
	// Write to current stream
    if( isWriteable( currentUrl ) )
    {
        // Return to where we were before
        int backAmount = -rBuf.amount();
        
        // Ensure buffer is flushed
        rBuf.init();
        
        // Move back then write
        fseek(fp, backAmount, SEEK_CUR);
        
        int ret = fwrite((void *)buffer, size, count, fp);
        
        if(ret == (int)count)
        {
            // Fine
            return ret;
        }
        else
        {
            // Set flag
            setFailed( currentUrl );
            
            // Not fine
            return ret;
        }
    }
    else
    {
        // Catch all other cases
        return 0;
	}
}

void File::close()
{
	// Erase filename
	currentUrl[0] = 0;
	
	// Don't do anything if this medium is failed
	if( hasFailed( currentUrl ) ) { return; }
	
	// Close any open handles
	switch(fType)
	{
		case STREAM_FILE:
			// Close file
			if(fp) { fclose(fp); }
			if(dir) { dirclose(dir); }
			
			// Ensure we don't close again
			fp = NULL;
			dir = NULL;
			
			break;
		default:
			break;
	}
	
	// Set file type to none
	fType = STREAM_NONE;
}

int File::seek(int32_t offset, int origin)
{
	// Don't do anything if this medium is failed
	if( hasFailed( currentUrl ) ) { return -1; }
	
	// Return to where we were before
	int backAmount = -rBuf.amount();
	
	// Ensure buffer is flushed
	rBuf.init();
	
	int ret;
	
	// Seek current stream
	switch(fType)
	{
		case STREAM_FILE:
		{
			// Move back then seek
			if(origin == SEEK_CUR) 
			{ 
				ret = fseek(fp, offset + backAmount, SEEK_CUR); 
			}
			else
			{
				ret = fseek(fp, offset, origin);
			}
			
			if(ret == 0)
			{
				// Fine
				return ret;
			}
			else
			{
				// Set flag
				setFailed( currentUrl );
				
				// Not fine
				return ret;
			}
			break;
		}
		default:
			// Catch all other cases
			return 0;
	}
}

uint32_t File::tell()
{
	// Don't do anything if this medium is failed
	if( hasFailed( currentUrl ) ) { return 0; }
	
	// Seek current stream
	switch(fType)
	{
		case STREAM_FILE:
			return ftell(fp) - rBuf.amount();
		default:
			// Catch all other cases
			return 0;
	}
}

int File::findNextFile(char *file, FileType &f)
{
	switch(fType)
	{
		case STREAM_FILE:
		{		
			struct stat st;
            
            // Don't do anything if this medium is failed
            if( hasFailed( currentUrl ) ) 
            { 
                if(listedBack) { return 0; }
                
                f = FILE_DIRECTORY;
                strcpy(file, "..");
                listedBack = true;
                
                return 1;
            }
            
			// Grab next directory entry
			int result = dirnext(dir,file,&st);
			
			if(!result)
			{
				// Normal
				if(st.st_mode & S_IFDIR)
				{
					if(strcmp(file, "..") == 0)
					{
						// Remember that we listed this
						listedBack = true;
					}
					
					// Directory
					f = FILE_DIRECTORY;
				}
				else
				{
					// Regular file
					f = FILE_NORMAL;
				}
				
				// Return normal
				return 1;
			}
			else
			{
				// Send back '..' entry if needed
				if(!listedBack)
				{
					f = FILE_DIRECTORY;
					strcpy(file, "..");
					listedBack = true;
					
					return 1;
				}
				
				// End, set null
				f = FILE_NONE;
				file[0] = 0;
				
				// Return end
				return 0;
			}
			
			break;
		}
		case STREAM_ROOT:
		{
			// If failed, don't list
			while(rootPos < filesystemcount)
			{
				if( !filesystems[rootPos].failed )
                {
                    // Found the next one
                    break; 
                }
				
				rootPos++;
			}
			
            if( rootPos < filesystemcount )
            {
                // Copy in next good filesystem
                strcpy( file, filesystems[rootPos].protocol );
                strcat( file, "://" );
                
                // All directories
                f = FILE_DIRECTORY;
            }
            else
            {
                // No more files
                file[0] = 0;
                f = FILE_NONE;
                return 0;
			}
			
			// Remember next entry
			rootPos++;
			
			return 1;
		}
		default:
			// End, set null
			f = FILE_NONE;
			file[0] = 0;
			
			return 0;
	}
}

int File::findNextString(char *str, int len)
{
	int outpos = 0;
	
	while(true)
	{
		// Grab next character
		char c;		
		
		if(read(&c, 1, 1) <= 0)
		{
			if(outpos == 0)
			{
				// No data left
				str[0] = 0;
				return -1;
			}
			
			// End of file, end of string
			str[outpos] = 0;
			return outpos;
		}
		
		switch(c)
		{
			case '\n':
				if(last_char == '\r')
				{
					// Ignore, windows ending
					last_char = 0;
					
					break;
				}
				
				// Fallthrough
			case '\r':
				// End of string!
				last_char = c;
				
				// Ensure properly null terminated
				str[outpos] = 0;				
				return outpos;
			default:
				// Just store
				str[outpos++] = c;
				
				break;
		}
		
		if(outpos >= len)
		{
			// Too long
			str[outpos] = 0;
			return outpos;
		}
	}
	
	return 0;
}

void File::getCurrentUrl(char *url)
{
	// Copy stored url
	strcpy(url, currentUrl);
}

void File::tryOpenFile( FileSystem *fs )
{
    if( !fs ) { return; }
    char name[256];
    
    // First, is it even inserted?
    if( !fs->interface->isInserted() )
    {
        // No, set failed and move on
        fs->revision++;
        fs->failed = true;
        return;
    }
    
    // Try twice, the first time through it could be not failed yet but had been
    // removed and reinserted.  So it will come back failed from the open call and
    // we need to try again.
    for( int i = 0; i < 2; i++ )
    {
        if( fs->failed )
        {
            strcpy( name, fs->protocol );
            strcat( name, ":" );
            
            // Shut down old handle
            fatUnmount( name );
            fs->interface->shutdown();
            
            // Keep track of the fact that this isn't "fresh" anymore
            fs->revision++;
            
            // Try to reopen it
            if( !(fs->interface->startup() && fatMountSimple(fs->protocol, fs->interface)) )
            {
                // Still failed, don't do anything
                fs->failed = true;
                break;
            }
        }
        
        // Try to enumerate the root directory
        strcpy( name, fs->protocol );
        strcat( name, ":/" );
        
        // The open call will set failed if this doesn't work
        fs->failed = false;
        open(name, "d");
        close();
        
        if( fs->failed == false )
        {
            // System is good, lets quit (potentially early)
            break;
        }
    }
}

void File::enumerateInterfaces(void)
{
    FileSystem *interface;
    
    // Shouldn't be doing this in more than one thread
    LWP_MutexLock(detectLock);
    
    // Try to find SD interface
    if( 0 == (interface = findFileSystem( "sd://" )) )
    {
        addInterface( "sd", sd );
        interface = findFileSystem( "sd://" );
        interface->writeable = true;
    }
    
    // Try to open it to see if it's failed
    tryOpenFile( interface );
    
    // Try to find USB interface
    if( 0 == (interface = findFileSystem( "usb://" )) )
    {
        addInterface( "usb", usb );
        interface = findFileSystem( "usb://" );
        interface->writeable = false;
    }
    
    // Try to open it to see if it's failed
    tryOpenFile( interface );
    
    // Try to enumerate any shares
    for( uint32_t i = 0; i < sharecount; i++ )
    {   
        // Try to find share interface
        char share[MAX_SHARE_LENGTH];
        
        strcpy( share, shares[i].name );
        strcat( share, "://" );
        
        if( 0 == (interface = findFileSystem( share )) )
        {
            addInterface( shares[i].name, NULL );
            smbInitDevice( shares[i].name, shares[i].user, shares[i].pass, shares[i].share, shares[i].location );
            interface = findFileSystem( share );
            interface->writeable = false;
        }
        
        // Try to open it to see if it's failed
        if( interface->failed )
        {
            smbClose( shares[i].name );
            smbInitDevice( shares[i].name, shares[i].user, shares[i].pass, shares[i].share, shares[i].location );
            
            // No checking now, too expensive, will happen when user tries browsing
            interface->failed = false;
        }
    }
    
    // Safe now!
    LWP_MutexUnlock(detectLock);
}

void File::enumerateShares()
{
    static bool done = false;
    
    // Forceably disallow multiple calls
    if( done ) { return; }
    
    // Open list of shares
    open(DATAPATH "shares/", "d");
    
    // Loop through and try to add each
    FileType ftype;
    char filename[MAXPATHLEN];
    
    while( findNextFile( filename, ftype ) )
    {
        if( ftype == FILE_NORMAL )
        {
            /* Try to open this file and parse it */
            char absolute[MAXPATHLEN];
            Share share;
            
            /* Construct a path */
            strcpy( absolute, DATAPATH "shares/" );
            strcat( absolute, filename );
            
            /* Can this be loaded? */
            if(share.loadShare( absolute )) { continue; }
            
            sharecount++;
            shares = (ShareList *)realloc( shares, sizeof( shares[0] ) * sharecount );
            
            share.getString("Name", shares[sharecount-1].name);
            share.getString("Location", shares[sharecount-1].location);
            share.getString("Share", shares[sharecount-1].share);
            share.getString("Username", shares[sharecount-1].user, "Guest");
            share.getString("Password", shares[sharecount-1].pass, "");
            
            /* @TODO make sure name is permissible (lowercase alphanum) */
        }
    }
    
    // Close list of shares
    close();
    
    // Don't allow second call
    done = true;
}

bool File::isValidURL(char *url)
{
	if(!url) { return false; }
	
	char tmp[9];
	
	strncpy(tmp, url, 8);
	tmp[8] = 0;
	
	strlwr(tmp);
	
	if(strncmp(tmp,"http://", 7) == 0)
	{
		return true;
	}
	
	if(strncmp(tmp,"https://", 8) == 0)
	{
		return true;
	}
	
	if(strncmp(tmp,"icy://", 8) == 0)
	{
		return true;
	}
    
    // Do we have it enumerated?
    if( findFileSystem( url ) )
    {
        return true;
    }
	
	return false;		
}

#ifdef DEBUG_MODE
int File::getCurrentParameter( int param )
{
    switch( param )
    {
        case 0:
            return readAheadAmount;
        case 1:
            return readAheadMax;
    }
    
    return 0;
}
#endif