#include <string.h>
#include <malloc.h>
#include "mp3source.h"

#define MP3_READ_SIZE 1024
#define MP3_INITIAL_READ_SIZE 8192

Mp3Source::Mp3Source() : SoundSource()
{
	// Ensure no bad open causes a false success on reading
	_length = 0;
	mh = NULL;
	
	// Init mpg123
	mpg123_init();
}

Mp3Source::~Mp3Source()
{
	// Kill file
	close();
	
	// Kill mpg123
	mpg123_exit();
}

OpenStatus Mp3Source::openLocation(File *sourceFile)
{
	// Remember file source object
	_sourceFile = sourceFile;
	
	// Get file size
	sourceFile->seek(0, SEEK_END);
	fileLength = sourceFile->tell();
	sourceFile->seek(0, SEEK_SET);
	
	// No id3v2 yet
	seek_start = 0;
	
	// Get bitrates
	const long *rates;
	size_t num_rates;
	
	mpg123_rates(&rates, &num_rates);
	
	// Get new mpg123 handle
	int error;
	mh = mpg123_new(NULL, &error);
	
	// Ensure we don't continue if something went wrong
	if(!mh) { return OPEN_FAILURE; }
	
	// Streaming mode
	mpg123_open_feed(mh);
	
	// Ensure that this output format will not change (it could, when we allow it).
	mpg123_format_none(mh);
	
	// Set all bitrates as ok
	for(int i = 0; i < (int)num_rates; ++i) 
	{
		mpg123_format(mh, rates[i], MPG123_STEREO, MPG123_ENC_SIGNED_16);
	}
	
	// Grab first frame to figure out parameters
	int result;
	
	// Allocate buffer
	unsigned char *buf = (unsigned char *)malloc(MP3_INITIAL_READ_SIZE);
	
	// Loop to get information
	do 
	{
		// Read in
		_sourceFile->read(buf, 1, MP3_INITIAL_READ_SIZE);
		
		// Seek past an ID3v2 header if present
		if(buf[0] == 'I' && buf[1] == 'D' && buf[2] == '3')
		{
			// Bingo
			u32 header_size = ((buf[6] & 0x7F) << 21) | ((buf[7] & 0x7F) << 14) | ((buf[8] & 0x7F) << 7) | (buf[9] & 0x7F);
			
			// Jump ahead past header
			_sourceFile->seek(10 + header_size, SEEK_SET);
			
			// Ensure no seeking to id3 header
			seek_start = 10 + header_size;
			
			// Read in
			_sourceFile->read(buf, 1, MP3_INITIAL_READ_SIZE);
		}
		
		// Feed to mpg123
		size_t fakegot;
		result = mpg123_decode(mh, buf, MP3_INITIAL_READ_SIZE, NULL, 0, &fakegot);
	} while(result == MPG123_NEED_MORE);
	
	// Free buffer
	free(buf);
	
	if(result != MPG123_NEW_FORMAT)
	{
		// Failed to get data		
		mpg123_close(mh);
		mh = NULL;
		
		// Exit out of here, no recovery
		return OPEN_FAILURE;
	}

	// Unneeded value encoding
	int encoding;
	
	// Grab frequency to play back as well as number of channels
	result = mpg123_getformat(mh, &frequency, &channels, &encoding);
	
	// Set file length
	mpg123_fake_set_length(mh, fileLength);
	
	// Grab length
	samples = mpg123_length(mh);
	
	// Return success
	return OPEN_SUCCESS;
}

int Mp3Source::renderChunk(uint8_t *outbuf, int &renderSamples)
{	
	// Clear data to ensure no garbage bytes
	memset(outbuf, 0, renderSamples * 4);
	
	// Start with assumption that we have enough data
	int result = MPG123_OK;
	
	// Remember end of file
	int over = 0;
	
	// Bookkeeping
	int need = renderSamples * channels * 2;
	int have_read = 0;
	
	// Ensure we don't read too much
	int read_times = 0;
	
	// Allocate buffer
	unsigned char *buf = (unsigned char *)malloc(MP3_READ_SIZE);
	
	// Loop, grabbing enough data to get samples
	do
	{
		// Bookkeeping
		int dataIn = 0;
		size_t have_now = 0;
		
		// Figure out if we need more data
		if(result == MPG123_NEED_MORE)
		{
			if(over)
			{
				// We exhausted our file
				free(buf);
				
				// Ensure we don't create garbage audio
				memset(outbuf + have_read, 0, need);
				
				// More data next time
				return 0;
			}
			
			// Read in
			dataIn = _sourceFile->read(buf, 1, MP3_READ_SIZE);
			read_times++;
			
			if(dataIn < MP3_READ_SIZE)
			{
				// End of file :<
				over = 1;
			}
		}
		
		// Grab data
		result = mpg123_decode(mh, buf, (size_t)dataIn, outbuf + have_read, need, &have_now);
		
		// Ensure we keep track of newly gotten data
		need -= have_now;
		have_read += have_now;
		
		// if we finished, then exit with success
		if(need == 0)
		{
			// Free buffer
			free(buf);
			
			// More data next time
			return 1;
		}
	} while((result == MPG123_NEED_MORE || result == MPG123_ERR) && read_times < 10);
	
	// Free buffer
	free(buf);
	
	// Abnormal exit
	return 0;
}

void Mp3Source::seekSeconds(int seconds)
{
	// Figure out seeking by seconds
	uint64_t seekoff = (int)((double)(fileLength - seek_start) * ((double)seconds / (double)(samples / frequency))) + seek_start;
	
	// Seek to requested location
	_sourceFile->seek(seekoff, SEEK_SET);
}

int Mp3Source::tellSeconds()
{
	// Calculate number of seconds in
	if(samples)
	{
		return (int)((double)(samples / frequency) * ((double)(_sourceFile->tell() - seek_start) / (double)(fileLength - seek_start)));
	}
	
	// No position
	return -1;
}

int Mp3Source::getLength()
{
	if(samples && frequency)
	{
		// Length in seconds to sound processor	
		return samples / frequency;
	}
	
	// No length
	return -1;
}

uint32_t Mp3Source::getFrequency()
{
	// Send back pre-fetched frequency of playback
	return frequency;
}

void Mp3Source::close()
{
	if(mh)
	{
		// Close the mpg123 handle
		mpg123_delete(mh);
		mh = NULL;
	}
}

int Mp3Source::supportsFormat(const char *ext)
{
	// Only one format
	if(strcmp(ext, "mp3") == 0) { return 1; }
	if(strcmp(ext, "mp2") == 0) { return 1; }
	if(strcmp(ext, "mp1") == 0) { return 1; }
	if(strcmp(ext, "mpa") == 0) { return 1; }
	
	// We support no other formats, so this function always fails
	return 0;
}

void Mp3Source::readAhead()
{
	// Read ahead some
	_sourceFile->readAhead();
}

