#include <string.h>
#include <malloc.h>

#include "sndhsource.h"

SndhSource::SndhSource() : SoundSource()
{
	// Set up emu
	buffer = NULL;
	sc68 = NULL;
}

SndhSource::~SndhSource()
{
	// Kill file
	close();
}

OpenStatus SndhSource::openLocation(File *sourceFile)
{
	// Try to open wav file
	_sourceFile = sourceFile;
	
	// Grab length
	sourceFile->seek(0, SEEK_END);
	uint32_t size = sourceFile->tell();
	sourceFile->seek(0, SEEK_SET);
	
	// Grab file to memory
	buffer = malloc(size);
	sourceFile->read(buffer, 1, size);
	
	// Set up sc68
	memset(&init68, 0, sizeof(init68));
	
	// Set dynamic handler (required).
	init68.alloc = malloc;
	init68.free = free;
	
	sc68 = api68_init(&init68);	
	
	if(!sc68)
	{
		// Library failed
		free(buffer);		
		buffer = NULL;
		
		return OPEN_FAILURE;
	}
	
	// Load an sc68 file.
	if(api68_load_mem(sc68, buffer, size) < 0) 
	{
		api68_shutdown(sc68);
		free(buffer);		
		buffer = NULL;
		
		return OPEN_FAILURE;
	}
	
	track_count = api68_tracks(sc68);
	
	// Set a track (optional)
	api68_play(sc68, 0, 0);
	
	// Return success
	return OPEN_SUCCESS;
}

int SndhSource::renderChunk(uint8_t *buf, int &renderSamples)
{
	// Ensure no blips at end
	memset(buf, 0, renderSamples * 4);
	
	// For easier assignment and manipulation
	unsigned short *tBuffer = (unsigned short *)buf;
	
	// Grab latest samples
	int code = api68_process(sc68, tBuffer, renderSamples);
	
	if(code & API68_END)
	{
		// No more data
		return 0;
	}
	
	// More data is available
	return 1;
}

void SndhSource::seekSeconds(int seconds)
{
	// No Seeking
	track = (seconds + 2) / 5;
	
	// Start the song initialize
	api68_play(sc68, track + 1, 0);
}

int SndhSource::tellSeconds()
{
	// Calculate number of seconds in
	return track * 5;
}

bool SndhSource::hasTime()
{
	// Disable seconds display
	return false;
}

int SndhSource::getLength()
{
	// Length in seconds to sound processor
	if(track_count < 2)
	{
		return -1;
	}
	else
	{
		// Fake seconds for seeking
		return (track_count - 1) * 5;
	}
}

uint32_t SndhSource::getFrequency()
{
	// Send back pre-fetched frequency of playback
	return init68.sampling_rate;
}

void SndhSource::close()
{
	api68_stop(sc68);
	api68_shutdown(sc68);
	
	sc68 = NULL;
	
	// Kill file
	if(buffer) { free(buffer); }
	
	buffer = NULL;
}

int SndhSource::supportsFormat(const char *ext)
{
	// Three formats
	if(strcmp(ext, "snd") == 0) { return 1; }
	if(strcmp(ext, "sndh") == 0) { return 1; }
	if(strcmp(ext, "sc68") == 0) { return 1; }
	
	// We support no other formats, so this function always fails
	return 0;
}
