#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogcsys.h>
#include <math.h>
#include <samplerate.h>

#include "common.h"
#include "sound.h"
#include "wavsource.h"
#include "modulesource.h"
#include "mp3source.h"
#include "vorbissource.h"
#include "flacsource.h"
#include "gmesource.h"
#include "sidsource.h"
#include "sndhsource.h"
#include "ymsource.h"

#define INPUT_SIZE              4096
#define MIN_OUTPUT_SIZE         4096
#define MAX_OUTPUT_SIZE         (MIN_OUTPUT_SIZE * 16)

#define FFT_SIZE                1024
#define FFT_DIVISOR             1

#define SOUND_RENDER_STACK      (256 * 1024)
#define SOUND_RENDER_PRIORITY   63

typedef struct
{
	u8 outBuffer[2][MIN_OUTPUT_SIZE];
	RingBuffer ringBuf;
	
	int fftResult[16];
    uint32_t fftHeartbeat;
	
	int currentBuffer;
    
    int *paused;
    FFT *fft;
} SOUND_BUFFER;

typedef union 
{
	struct 
	{
		u16 hi;
		u16 lo;
	} aword;
	u32 adword;
} dword;

static SOUND_BUFFER soundBuf ATTRIBUTE_ALIGN(32);
static int restartFile = 0;

// Mutex for referencing the current rendering driver
mutex_t renderLock;

// We want multiple accessor classes to the same sound driver
int Sound::instance = 0;
	
File Sound::sourceFile;
SoundSource *Sound::soundSource = NULL;
SoundSource **Sound::availableSources = NULL;

FFT Sound::fft(FFT_SIZE);
Light Sound::light;
bool Sound::_lenabled = false;

int Sound::started;
int Sound::finished;
int Sound::paused;
int Sound::naturalFinish;
int Sound::loaded;

// For threads
void *soundRenderThread( void * args );

static void getExtension(const char *url, char *ext, int max)
{
	// Grab extensions
	char *tExt = strdup( url );
    strcpy( tExt, "" );
	
	// Only look a little ways back
	for(int i = 0; i < (max - 1); i++)
	{
		if(url[strlen(url) - i] == '.')
		{
			// Found it!
			strcpy(tExt, &url[strlen(url) - i] + 1);
			
			break;
		}
	}
	
	// Lowercase
	strlwr(tExt);
	
	// Copy out
	strcpy(ext, tExt);
    
    // Free duplicated memory
    free( tExt );
}

static void DMACallBack()
{
	// Start DMA transfer of next sound chunk
	AUDIO_StopDMA();
	
	// Here too for stuttering in skin changes
	if(soundBuf.paused && *soundBuf.paused)
	{
		// Sound should be empty
		memset(soundBuf.outBuffer[soundBuf.currentBuffer], 0, MIN_OUTPUT_SIZE);
	}
	
    // Make sure range is flushed and send it out
    DCFlushRange(soundBuf.outBuffer[soundBuf.currentBuffer], MIN_OUTPUT_SIZE);
	AUDIO_InitDMA((u32)soundBuf.outBuffer[soundBuf.currentBuffer], MIN_OUTPUT_SIZE);
   
	AUDIO_StartDMA();

	// Swap buffers for next fill
	soundBuf.currentBuffer ^= 1;
    
    // Size up next copy
    int dataToCopy = soundBuf.ringBuf.amount();
    
    // Only copy out what we have ready
    if( dataToCopy > MIN_OUTPUT_SIZE ) { dataToCopy = MIN_OUTPUT_SIZE; }
    
    // Copy out the data
    soundBuf.ringBuf.get(soundBuf.outBuffer[soundBuf.currentBuffer], dataToCopy);
    
    // Fill in the rest with nothing
    if( dataToCopy < MIN_OUTPUT_SIZE )
    {
        memset( soundBuf.outBuffer[soundBuf.currentBuffer] + dataToCopy, 0, MIN_OUTPUT_SIZE - dataToCopy );
    }
    
    // Make sure that FFT code can update sometimes
    soundBuf.fftHeartbeat++;
}

static void Resample(s16 *samples, int numSamples, u32 src_samplerate)
{
    // local context for resampling
    static u32 cursamplerate = 0;
    static SRC_STATE *leftchannel = 0;
    static SRC_STATE *rightchannel = 0;
    
    // calculate current ratio
    double ratio = ((double)48000) / ((double)src_samplerate);
    int dstlen = (numSamples * ratio) + 1000;
    
    if( cursamplerate != src_samplerate || restartFile )
    {
        // do we need to close old context?
        if( leftchannel && rightchannel )
        {
            leftchannel = src_delete( leftchannel );
            rightchannel = src_delete( rightchannel );
        }
        
        // Create a context to operate in
        int error;
        leftchannel = src_new(SRC_SINC_FASTEST, 1, &error);
        rightchannel = src_new(SRC_SINC_FASTEST, 1, &error);
        
        // remember that we switched
        cursamplerate = src_samplerate;
        
        // Remember that we restarted
        restartFile = 0;
    }
    
    // Room for temporary
    float *inleft = (float*)malloc( sizeof( float ) * numSamples );
    float *inright = (float*)malloc( sizeof( float ) * numSamples );
    float *outleft = (float*)malloc( sizeof( float ) * dstlen );
    float *outright = (float*)malloc( sizeof( float ) * dstlen );
    
    if( inleft && inright && outleft && outright )
    {    
        // Convert samples to float
        for( int i = 0; i < numSamples; i++ )
        {
            inleft[i] = (float)samples[(i << 1)] / (float)32768;
            inright[i] = (float)samples[(i << 1) + 1] / (float)32768;
        }
        
        // Resampling seems to be finnicky
        uint32_t state = IRQ_Disable();
        
        // Resample
        int count;
        {
            SRC_DATA left;
            SRC_DATA right;
            
            /* Set up in/out */
            left.data_in = inleft;
            left.data_out = outleft;
            right.data_in = inright;
            right.data_out = outright;
            
            /* Set up sample counts */
            left.input_frames = numSamples;
            left.output_frames = dstlen;
            right.input_frames = numSamples;
            right.output_frames = dstlen;
            
            /* Fake */
            left.end_of_input = 0;
            right.end_of_input = 0;
            
            /* Set up ratio for conversion */
            left.src_ratio = ratio;
            right.src_ratio = ratio;
            
            /* Resample! */
            src_process(leftchannel, &left);
            src_process(rightchannel, &right);
            
            // take max (should be same, but still)
            count = MAX(left.output_frames_gen, right.output_frames_gen);
        }
        
        // Safe now
        IRQ_Restore(state);
        
        // Clipping handling
        float maxclip = 1.0;
        for( int i = 0; i < count; i++ )
        {
            /* Left side */
            float tclip = fabs(outleft[i]);
            maxclip = MAX(tclip, maxclip);
            
            /* Right side */
            tclip = fabs(outright[i]);
            maxclip = MAX(tclip, maxclip);
        }
        
        // Convert and place into ringbuffer
        for( int i = 0; i < count; i++ )
        {
            s16 val;
            
            val = (s16)((outleft[i] / maxclip) * (float)32767);
            soundBuf.ringBuf.put(&val, sizeof(val));
            
            val = (s16)((outright[i] / maxclip) * (float)32767);
            soundBuf.ringBuf.put(&val, sizeof(val));
        }
    }
    else
    {
        /* Couldn't convert */
        for( int i = 0; i < numSamples * 2; i++ )
        {
            uint16_t out = 0;
            soundBuf.ringBuf.put(&out, sizeof(out));
        }
    }
    
    // Free memory
    free( inleft );
    free( inright );
    free( outleft );
    free( outright );
}

Sound::Sound()
{
    // Get current instance
    int curInstance = instance++;
    
	// Instance tracking
	if(!curInstance)
	{
        // Initialize mutex for reading ahead
        LWP_MutexInit(&renderLock, false);
        
		// Set buffer defaults
		soundBuf.currentBuffer = 0;
        soundBuf.fftHeartbeat = 0;
		
		// Remember that the finish was natural or not
		naturalFinish = 0;
		
		// Remember for dma callback
		soundBuf.paused = &paused;
        soundBuf.fft = &fft;
        
		// Initialize audio
		AUDIO_Init(NULL);
		AUDIO_SetDSPSampleRate(AI_SAMPLERATE_48KHZ);
		
        // Start for the first time
		AUDIO_RegisterDMACallback(DMACallBack);
        
		// Force a DMA update
		DMACallBack();
        
		// Create audio sources (this is the only place with any knowledge of the actual file source names)
		availableSources = (SoundSource **)malloc(10 * sizeof(SoundSource *));
		
		// Add each to list
		availableSources[0] = new WavSource();
		availableSources[1] = new ModuleSource();
		availableSources[2] = new Mp3Source();
		availableSources[3] = new VorbisSource();
		availableSources[4] = new FlacSource();
		availableSources[5] = new GmeSource();
		availableSources[6] = new SidSource();
		availableSources[7] = new SndhSource();
		availableSources[8] = new YmSource();
		availableSources[9] = NULL;
        
        // Create thread for rendering
        static lwp_t id;
        LWP_CreateThread(&id, soundRenderThread, NULL, NULL, SOUND_RENDER_STACK, SOUND_RENDER_PRIORITY);
	}
}

Sound::~Sound()
{
	// Instance tracking
	instance--;
	
	if(!instance)
	{   
		// Kill all of the audio renderers
		int i = 0;
		
		while(availableSources[i])
		{
			// Remove source from memory
			delete availableSources[i];
			
			// Go to next source
			i++;
		}
		
		// Ensure container is freed too
		free(availableSources);
		availableSources = NULL;
        
        // Make sure if for some ungodly reason we get a new sound object we don't fuck up
        LWP_MutexDestroy(renderLock);
	}
}

void Sound::updateBuffers()
{
	// Check for update needed
	if(soundBuf.ringBuf.amount() > MAX_OUTPUT_SIZE)
	{		
		return;
	}
    
    LWP_MutexLock(renderLock);

	if(paused)
	{	
		if(finished) { paused = 0; }
	}
    else if(finished > 0)
	{
		// Finished with sound
		if(finished == 1)
		{
			// Update the state
			finished = 2;
		}
		else if(finished == 2)
		{
			// Set done state
			finished = 3;
			paused = 0;
			started = 0;
		}
	}
    
    LWP_MutexUnlock(renderLock);
	
	// Fill current buffer with next bit of audio
	if( getSoundState() == SOUND_PLAYING ) 
	{ 
		// Loop until no more data available (if we hit the end first before finishing the resample)		
		int result = 1;
		
		// Loop and add samples after resampling
		while(soundBuf.ringBuf.amount() < MAX_OUTPUT_SIZE && result)
		{
			// Grab samples
			s16 samples[INPUT_SIZE];
			int numSamples = (sizeof( samples ) / sizeof( samples[0] )) / 2;
            
            if( getSoundState() == SOUND_PLAYING )
            {
                // Cant switch audio modes WHILE rendering
                LWP_MutexLock(renderLock);
                
                // Try to render result
                result = soundSource->renderChunk((uint8_t *)samples, numSamples);
                
                // Resample to add to ring buffer
                Resample(samples, numSamples, soundSource->getFrequency());
                
                LWP_MutexUnlock(renderLock);
            }
            else
            {
                // exit out, sound finished
                break;
            }
		}
		
		if(!result)
		{
            // Don't update these without permission
            LWP_MutexLock(renderLock);
            
			// Remember end state
			finished = 1;
			naturalFinish = 1;
            
            // Relinquish
            LWP_MutexUnlock(renderLock);
		}
	}
}

void *soundRenderThread( void * args )
{
    Sound sound;
    
    while( 1 )
    {
        // Simply render and yield
        sound.updateBuffers();
        LWP_YieldThread();
    }
    
    return NULL;
}

void Sound::update()
{
    // Nothing needs to be updated yet
    if(soundSource) { soundSource->readAhead(); }
}

int Sound::finishedNaturally()
{
    int ret = 0;
    
	// Only return if we did actually finish
	if(getSoundState() == SOUND_STOPPED)
	{
        LWP_MutexLock(renderLock);
        
		if(naturalFinish)
		{
			// Yup
			ret = 1;
		}
        
        LWP_MutexUnlock(renderLock);
	}
	
	// Not finished or no natural finish?
	return ret;
}

void Sound::resetNaturalFinish()
{
    LWP_MutexLock(renderLock);
    
	// To control from filelist to assure we handled the event
	naturalFinish = 0;
    
    LWP_MutexUnlock(renderLock);
}

int Sound::playsFormat(const char *url)
{
	// Get extension
	char ext[10];
	getExtension(url, ext, 10);
	
	// Loop through available sources and find one that works
	int i = 0;
	
	while(availableSources[i])
	{
		// Test to see if it supports it
		if(availableSources[i]->supportsFormat(ext)) { return 1; }
		
		// Go to next source
		i++;
	}
	
	// Couldn't find a valid source
	return 0;
}

int Sound::loadSound(const char *url)
{	
    LWP_MutexLock(renderLock);
    
	// Remember that the finish was natural or not
	naturalFinish = 0;
	
	// Grab extensions
	char ext[10];
	getExtension(url, ext, 10);
    
	// Temporary sound source
	SoundSource *tSoundSource = NULL;
	
	// Loop through available sources and find one that works
	int i = 0;
	
	while(availableSources[i])
	{
		// Test to see if it supports it
		if(availableSources[i]->supportsFormat(ext))
		{ 
			// Remember the working source
			tSoundSource = availableSources[i];
			
			// Exit out of loop
			break;
		}
		
		// Go to next source
		i++;
	}
    
    // Close now as closeSound needs to use the mutex
    LWP_MutexUnlock(renderLock);
	
	if(!tSoundSource)
	{
        // No renderer
		return 0;
	}
    
	// Close old sound
	closeSound();
	
	// Assign new buffer
	soundSource = tSoundSource;
	
    // Reaquire lock
    LWP_MutexLock(renderLock);
    
	// Open file
	if(sourceFile.open(url, "rb"))
	{
        // Clear out anything rendered
        clearData();
        
		// No sound source
		soundSource = NULL;
		
        // Bad renderer
        LWP_MutexUnlock(renderLock);
        
		// Stop current
		stopSoundInternal();
		
		return 0;
	}
	
	// Pass opened file to the sound source
	bool found = false;
	
	while(!found)
	{
		switch(soundSource->openLocation(&sourceFile))
		{
			case OPEN_FAILURE:            
                // Bad renderer
                LWP_MutexUnlock(renderLock);
                
				// Error
				closeSound();
				
				return 0;
			case OPEN_SUCCESS:
				// Opened!
				found = true;
				
				break;
			case OPEN_TRYAGAIN:                
                // Bad renderer
                LWP_MutexUnlock(renderLock);
                
				// Unhandled
				return 0;
			case OPEN_ANOTHERDECODER:
				// Find new decoder from extension
				char tExt[128];				
				soundSource->getFormat(tExt);
				
				// Ensure no false positives
				soundSource = NULL;
				
				int i = 0;				
				while(availableSources[i])
				{
					// Test to see if it supports it
					if(availableSources[i]->supportsFormat(tExt))
					{ 
						// Remember the working source
						tSoundSource = availableSources[i];
						
						// Exit out of loop
						break;
					}
					
					// Go to next source
					i++;
				}
				
				if(!tSoundSource)
				{
                    // Bad renderer
                    LWP_MutexUnlock(renderLock);
                    
					// No renderer
					return 0;
				}
				
				// Set renderer and retry loop
				soundSource = tSoundSource;
				break;
		}
	}
	
	// Initialize output ring buffer
	soundBuf.ringBuf.init();
	
	// Set defaults
	started = 0;
	paused = 0;
	finished = 0;
	loaded = 1;
    
    // force sound thread to redo the upsampling
    restartFile = 1;
    
    // Good to go
    LWP_MutexUnlock(renderLock);
	
	return 1;
}

void Sound::closeSound()
{
	// Stop current
	stopSoundInternal();
	
	// Kill current audio if needed (but don't delete)
	if(soundSource) 
	{
		loaded = 0;
		soundSource->close();
	}
	
	// Close audio file
	sourceFile.close();
}

void Sound::playSound()
{
	if(!soundSource) { return; }
    
    LWP_MutexLock(renderLock);
	
	if(!started || finished)
	{
		// Load if not loaded
		if(!loaded) 
		{
			soundSource->openLocation(&sourceFile);
			loaded = 1;
		}
		
        // Clear out anything rendered
        clearData();
        
		// Mark started
		started = 1;
		finished = 0;
		
		// Remember that the finish was natural or not
		naturalFinish = 0;
        
		// Ensure light doesn't stay on
		light.resetAverage();
	}
	else if(paused)
	{
		// Unpause
		paused = 0;
		
		// Ensure light doesn't stay on
		light.resetAverage();
	}
    
    LWP_MutexUnlock(renderLock);
}

void Sound::stopSound()
{
	stopSoundInternal();
	
	// Restart file
	if(soundSource)
	{
		sourceFile.seek(0, SEEK_SET);
		soundSource->close();
		
		loaded = 0;
	}
}

void Sound::clearData()
{
    // Clear out anything rendered
    soundBuf.ringBuf.clear();
    
    for( int i = 0; i < MIN_OUTPUT_SIZE; i++ )
    {
        // So we don't get blips
        uint16_t data = 0;        
        soundBuf.ringBuf.put(&data, sizeof( data ));
    }
}

void Sound::stopSoundInternal()
{
    // Clear out anything rendered
    clearData();
	
	// Ensure light doesn't stay on
	light.resetAverage();
    
    LWP_MutexLock(renderLock);
	
	// Set flags to not paused, not started
	started = 0;
	paused = 0;
	finished = 1;
    
    LWP_MutexUnlock(renderLock);
}

void Sound::pauseSound()
{
    LWP_MutexLock(renderLock);
	
	// Pause
	paused ^= 1;
	
    LWP_MutexUnlock(renderLock);
    
	// Ensure light doesn't stay on
	light.resetAverage();
}

int Sound::getSoundLength()
{
	// Ensure no null dereference
	if(!soundSource || !started) { return -1; }
	
	return soundSource->getLength();
}

bool Sound::getTimeEnabled()
{
	// Ensure no null dereference
	if(!soundSource) { return false; }
	
	return soundSource->hasTime();
}

int Sound::getSoundPosition()
{
	// Ensure no null dereference
	if(!soundSource || !started) { return -1; }
	
	// Figure out actual length
	int len = soundSource->tellSeconds();
	
	// Always return -1 if it's the case
	if(len < 0) { return len; }
	
	// Return 0 if sound is stopped
	if(getSoundState() == SOUND_STOPPED) { return 0; }
	
	// Return actual value otherwise
	return len;
}

void Sound::setSoundPosition(int pos)
{
	// Ensure no null dereference
	if(!soundSource) { return; }
	
	soundSource->seekSeconds(pos);
}

void Sound::getNowPlaying(char *str)
{
	// Ensure no null dereference
	if(!soundSource) 
	{ 
		// No bad strings
		str[0] = 0;
		
		return; 
	}
	
	// Return current song (fix this to pull from global format options)
	soundSource->getNowPlaying(str);
}

SoundState Sound::getSoundState()
{
    SoundState ret;

    /* Want to guarantee that nothing changes while we are returning this */
    LWP_MutexLock(renderLock);

	if(!soundSource)
	{
		// Sound isn't loaded
		ret = SOUND_UNLOADED;
	}
	else if(!started)
	{
		// Not started
		ret = SOUND_STOPPED;
	}
	else if(paused)
	{
		// It's paused
		ret = SOUND_PAUSED;
	}
	else 
    {
        // Its playing
        ret = SOUND_PLAYING;
    }
    
    /* Safe now */
    LWP_MutexUnlock(renderLock);
    
	return ret;
}

void Sound::getFFTResult(int *out)
{
    // Keep track of when we last updated
    static uint32_t last_update = (uint32_t)-1;
    
    if( last_update != (soundBuf.fftHeartbeat / FFT_DIVISOR) )
    {
        // FFT for current fill
        soundBuf.fft->setAudioData((int16_t *)soundBuf.outBuffer[soundBuf.currentBuffer ^ 1]);
        soundBuf.fft->getFFT(soundBuf.fftResult);
        
        // Remember we did this
        last_update = soundBuf.fftHeartbeat / FFT_DIVISOR;
    }
    
	// Use data for light
	light.setLightData(soundBuf.fftResult);
	
    // Copy it
    memcpy(out, soundBuf.fftResult, 16 * sizeof(int));
}

void Sound::enableDiskLight(bool enabled)
{
	// Pass on settings to fft which manages wii light
	light.setFlashEnabled(enabled);
	
	// Remember
	_lenabled = enabled;
}

bool Sound::getDiskLightEnabled()
{
	// return light status
	return _lenabled;
}
 
#ifdef DEBUG_MODE
int Sound::getCurrentParameter( int param )
{
    switch( param )
    {
        case 0:
            return soundBuf.currentBuffer;
        case 1:
            return MIN_OUTPUT_SIZE;
        case 2:
            return soundBuf.ringBuf.amount();
        case 3:
            return MAX_OUTPUT_SIZE;
    }
    
    return 0;
}
#endif
